@@ -540,8 +540,8 @@ fprintf(stderr, " '%s'\n", Tmp_win->name);
       old_wmhints_flags = Tmp_win->wmhints->flags;
       XFree ((char *) Tmp_win->wmhints);
     }
-    Tmp_win->wmhints = XGetWMHints(dpy, Event.xany.window);
-    if(Tmp_win->wmhints == NULL)
+    setup_wm_hints(Tmp_win);
+    if (Tmp_win->wmhints == NULL)
     {
       return;
     }
@@ -890,10 +890,17 @@ void HandleExpose(void)
 {
   XRectangle r;
 
-  if (Event.xexpose.count != 0)
+#if 0
+  /* This doesn't work well. Sometimes, the expose count is zero although
+   * dozens of expose events are pending.  This happens all the time during
+   * a shading animation.  SImply flush expose events unconditionally. */
+  if (Event.xexpose.count != 0||1)
   {
     flush_accumulate_expose(Event.xexpose.window, &Event);
   }
+#else
+  flush_accumulate_expose(Event.xexpose.window, &Event);
+#endif
   r.x = Event.xexpose.x;
   r.y = Event.xexpose.y;
   r.width = Event.xexpose.width;
@@ -1017,11 +1024,16 @@ void HandleMapRequestKeepRaised(Window KeepRaised, FvwmWindow *ReuseWin)
 
   if (ReuseWin == NULL)
   {
-    if(XFindContext(dpy, Event.xany.window, FvwmContext,
-		    (caddr_t *)&Tmp_win)==XCNOENT)
+    if (XFindContext(dpy, Event.xany.window, FvwmContext,
+		     (caddr_t *)&Tmp_win)==XCNOENT)
     {
       Tmp_win = NULL;
     }
+    else if (IS_MAP_PENDING(Tmp_win))
+    {
+      /* The window is already going to be mapped, no need to do that twice */
+      return;
+    }
   }
   else
   {
@@ -1067,15 +1079,15 @@ void HandleMapRequestKeepRaised(Window KeepRaised, FvwmWindow *ReuseWin)
   {
     int state;
 
-    if(Tmp_win->wmhints && (Tmp_win->wmhints->flags & StateHint))
+    if (Tmp_win->wmhints && (Tmp_win->wmhints->flags & StateHint))
       state = Tmp_win->wmhints->initial_state;
     else
       state = NormalState;
 
-    if(DO_START_ICONIC(Tmp_win))
+    if (DO_START_ICONIC(Tmp_win))
       state = IconicState;
 
-    if(isIconicState != DontCareState)
+    if (isIconicState != DontCareState)
       state = isIconicState;
 
     switch (state)
@@ -1324,6 +1336,26 @@ void HandleMapNotify(void)
 }
 
 
+Bool check_map_request(Display *display, XEvent *event, char *arg)
+{
+  unsigned long *rc;
+
+  if (event->type != MapRequest)
+  {
+    return False;
+  }
+  if (event->xmaprequest.window != *(unsigned long *)arg)
+  {
+    return False;
+  }
+  rc = (unsigned long *)arg;
+  (*arg)++;
+
+  /* Yes, it is correct that this function always returns False.  See
+   * comment below. */
+  return False;
+}
+
 /***********************************************************************
  *
  *  Procedure:
@@ -1375,8 +1407,34 @@ void HandleUnmapNotify(void)
   }
   if (must_return)
     return;
-  if(weMustUnmap)
-    XUnmapWindow(dpy, Event.xunmap.window);
+
+  if (weMustUnmap)
+  {
+    unsigned long win = (unsigned long)Event.xunmap.window;
+    Bool is_map_request_pending;
+
+    /* Using XCheckTypedWindowEvent() does not work here.  I don't have the
+     * slightest idea why, but using XCheckIfEvent() with the appropriate
+     * predicate procedure works fine. */
+    XCheckIfEvent(dpy, &dummy, check_map_request, (char *)&win);
+    /* Unfortunately, there is no procedure in X that simply tests if an
+     * event of a certain type in on the queue without waiting and without
+     * removing it from the queue.  XCheck...Event() does not wait but
+     * removes the event while XPeek...() does not remove the event but
+     * waits.  Now, this is a *real* hack:
+     *
+     * The predicate procedure used in the call always returns False.  Thus,
+     * no event is ever removed from the queue.  But when it is called with
+     * the correct event type and window, it adds one to the unsigned long
+     * pointer that was passed in as its last argument.  Of course this is
+     * using totally undocumented 'features' and may be very inefficient
+     * (im many events are pending). */
+    is_map_request_pending = (win != (unsigned long)Event.xunmap.window);
+    if (!is_map_request_pending)
+    {
+      XUnmapWindow(dpy, Event.xunmap.window);
+    }
+  }
   if(Tmp_win ==  Scr.Hilite)
   {
     Scr.Hilite = NULL;
@@ -1803,6 +1861,18 @@ void HandleEnterNotify(void)
   {
     return;
   }
+  else if (ewp->mode == NotifyNormal)
+  {
+    if (ewp->detail == NotifyNonlinearVirtual && ewp->focus == False &&
+	ewp->subwindow != None)
+    {
+      /* This takes care of some buggy apps that forget that one of their
+       * dialog subwindows has the focus after popping up a selection list
+       * several times (ddd, netscape). I'm not convinced that this does not
+       * break something else. */
+      refresh_focus(Tmp_win);
+    }
+  }
   else if (ewp->mode == NotifyUngrab)
   {
     /* Ignore events generated by grabbing or ungrabbing the pointer.  However,
@@ -1814,9 +1884,18 @@ void HandleEnterNotify(void)
       is_initial_ungrab_pending = False;
       xcrossing_last_grab_window = NULL;
     }
-    else if (Tmp_win && Tmp_win == xcrossing_last_grab_window)
+    else
     {
-      return;
+      if (ewp->detail == NotifyNonlinearVirtual && ewp->focus == False &&
+	  ewp->subwindow != None)
+      {
+	/* see comment above */
+	refresh_focus(Tmp_win);
+      }
+      if (Tmp_win && Tmp_win == xcrossing_last_grab_window)
+      {
+	return;
+      }
     }
   }
   if (Tmp_win)
