@@ -30,7 +30,9 @@
 #include "screen.h"
 #include "defaults.h"
 #include "module_interface.h"
+#include "focus.h"
 #include "stack.h"
+#include "events.h"
 #include "borders.h"
 #include "virtual.h"
 #include "gnome.h"
@@ -39,20 +41,160 @@
 
 static void RaiseOrLowerWindow(
   FvwmWindow *t, Bool do_lower, Bool allow_recursion, Bool is_new_window);
+#if 0
 static void ResyncFvwmStackRing(void);
+#endif
 static void ResyncXStackingOrder(void);
+static void BroadcastRestack(FvwmWindow *s1, FvwmWindow *s2);
+
+#define DEBUG_STACK_RING 1
+#ifdef DEBUG_STACK_RING
+/* debugging function */
+static void dump_stack_ring(void)
+{
+  FvwmWindow *t1;
+
+  if (!debugging_stack_ring)
+    return;
+  XBell(dpy, 0);
+  fprintf(stderr,"dumping stack ring:\n");
+  for (t1 = Scr.FvwmRoot.stack_next; t1 != &Scr.FvwmRoot; t1 = t1->stack_next)
+  {
+    fprintf(stderr,"    l=%d fw=0x%08x f=0x%08x '%s'\n", t1->layer,
+	    (int)t1, (int)t1->frame, t1->name);
+  }
+
+  return;
+}
+
+/* debugging function */
+void verify_stack_ring_consistency(void)
+{
+  Window root, parent, *children;
+  unsigned int nchildren, i;
+  FvwmWindow *t1, *t2;
+  int last_layer;
+  int last_index;
+
+  if (!debugging_stack_ring)
+    return;
+  XSync(dpy, 0);
+  t2 = Scr.FvwmRoot.stack_next;
+  if (t2 == &Scr.FvwmRoot)
+    return;
+  last_layer = t2->layer;
+
+  for (t1 = t2->stack_next; t1 != &Scr.FvwmRoot; t2 = t1, t1 = t1->stack_next)
+  {
+    if (t1->layer > last_layer)
+    {
+      fprintf(
+	stderr,
+	"vsrc: stack ring is corrupt! "
+	"'%s' (layer %d) is above '%s' (layer %d/%d)\n",
+	t1->name, t1->layer, t2->name, t2->layer, last_layer);
+      dump_stack_ring();
+      return;
+    }
+    last_layer = t1->layer;
+  }
+  t2 = &Scr.FvwmRoot;
+  for (t1 = t2->stack_next; t1 != &Scr.FvwmRoot; t2 = t1, t1 = t1->stack_next)
+  {
+    if (t1->stack_prev != t2)
+      break;
+  }
+  if (t1 != &Scr.FvwmRoot || t1->stack_prev != t2)
+  {
+    fprintf(
+      stderr,
+      "vsrc: stack ring is corrupt - fvwm will probably crash! "
+      "0x%08x -> 0x%08x but 0x%08x <- 0x%08x\n",
+      (int)t2, (int)t1, (int)(t1->stack_prev), (int)t1);
+    dump_stack_ring();
+    return;
+  }
+  MyXGrabServer(dpy);
+  if (!XQueryTree(dpy, Scr.Root, &root, &parent, &children, &nchildren))
+  {
+    MyXUngrabServer(dpy);
+    return;
+  }
+
+  last_index = nchildren;
+  for (t1 = Scr.FvwmRoot.stack_next; t1 != &Scr.FvwmRoot; t1 = t1->stack_next)
+  {
+    /* find window in window list */
+    for (i = 0; i < nchildren && t1->frame != children[i]; i++)
+      ;
+    if (i == nchildren)
+    {
+      fprintf(stderr,"vsrc: window already died: fw=0x%08x w=0x%08x '%s'\n",
+	      (int)t1, (int)t1->frame, t1->name);
+    }
+    else if (i >= last_index)
+    {
+      fprintf(
+	stderr, "vsrc: window is at wrong position in stack ring: "
+	"fw=0x%08x f=0x%08x '%s'\n", (int)t1, (int)t1->frame, t1->name);
+      dump_stack_ring();
+      fprintf(stderr,"dumping X stacking order:\n");
+      for (i = nchildren; i-- > 0; )
+      {
+	for (t1 = Scr.FvwmRoot.stack_next; t1 != &Scr.FvwmRoot;
+	     t1 = t1->stack_next)
+	{
+	  /* only dump frame windows */
+	  if (t1->frame == children[i])
+	  {
+	    fprintf(stderr,"  f=0x%08x\n", (int)children[i]);
+	    break;
+	  }
+	}
+      }
+      MyXUngrabServer(dpy);
+      XFree(children);
+      return;
+    }
+    last_index = i;
+  }
+  MyXUngrabServer(dpy);
+  XFree(children);
+
+  return;
+}
+#endif
 
 /* Remove a window from the stack ring */
 void remove_window_from_stack_ring(FvwmWindow *t)
 {
+  if (IS_SCHEDULED_FOR_DESTROY(t))
+  {
+    return;
+  }
   t->stack_prev->stack_next = t->stack_next;
   t->stack_next->stack_prev = t->stack_prev;
+  /* not really necessary, but gives a little more saftey */
+  t->stack_prev = NULL;
+  t->stack_next = NULL;
   return;
 }
 
-/* Add window t to the stack ring after window t2 */
+/* Add window t to the stack ring after window t */
 void add_window_to_stack_ring_after(FvwmWindow *t, FvwmWindow *add_after_win)
 {
+  if (IS_SCHEDULED_FOR_DESTROY(t))
+  {
+    return;
+  }
+  if (t == add_after_win || t == add_after_win->stack_next)
+  {
+    /* tried to add the window before or after itself */
+    fvwm_msg(ERR, "add_window_to_stack_ring_after",
+	     "BUG: tried to add window '%s' %s itself in stack ring\n", t->name,
+	     (t == add_after_win) ? "after" : "before");
+    return;
+  }
   t->stack_next = add_after_win->stack_next;
   add_after_win->stack_next->stack_prev = t;
   t->stack_prev = add_after_win;
@@ -60,6 +202,17 @@ void add_window_to_stack_ring_after(FvwmWindow *t, FvwmWindow *add_after_win)
   return;
 }
 
+/* Add a whole ring of windows. The list_head itself will not be added. */
+static void add_windowlist_to_stack_ring_after(
+  FvwmWindow *list_head, FvwmWindow *add_after_win)
+{
+  add_after_win->stack_next->stack_prev = list_head->stack_prev;
+  list_head->stack_prev->stack_next = add_after_win->stack_next;
+  add_after_win->stack_next = list_head->stack_next;
+  list_head->stack_next->stack_prev = add_after_win;
+  return;
+}
+
 FvwmWindow *get_next_window_in_stack_ring(FvwmWindow *t)
 {
   return t->stack_next;
@@ -70,6 +223,39 @@ FvwmWindow *get_prev_window_in_stack_ring(FvwmWindow *t)
   return t->stack_prev;
 }
 
+FvwmWindow *get_transientfor_fvwmwindow(FvwmWindow *t)
+{
+  FvwmWindow *s;
+
+  if (!t || !IS_TRANSIENT(t) || t->transientfor == Scr.Root ||
+      t->transientfor == None)
+    return NULL;
+  for (s = Scr.FvwmRoot.next; s != NULL; s = s->next)
+  {
+    if (s->w == t->transientfor)
+    {
+      return (s == t) ? NULL : s;
+    }
+  }
+
+  return NULL;
+}
+
+static FvwmWindow *get_transientfor_top_fvwmwindow(FvwmWindow *t)
+{
+  FvwmWindow *s;
+
+  s = t;
+  while (s && IS_TRANSIENT(s) && DO_STACK_TRANSIENT_PARENT(s))
+  {
+    s = get_transientfor_fvwmwindow(s);
+    if (s)
+      t = s;
+  }
+
+  return t;
+}
+
 /* Takes a window from the top of the stack ring and puts it at the appropriate
  * place. Called when new windows are created. */
 Bool position_new_window_in_stack_ring(FvwmWindow *t, Bool do_lower)
@@ -99,100 +285,88 @@ static void raise_over_unmanaged(FvwmWindow *t)
   Window *tops;
   int i;
   unsigned int num;
-  Bool found = False;
-  Window OR_Above  =  0;
-  Window topwin;
+  Window OR_Above = None;
   Window *wins;
   int count = 0;
-  FvwmWindow  *t2       =  NULL;
-  FvwmWindow  *junkwin  =  NULL;
+  FvwmWindow *t2 = NULL;
   unsigned int flags;
   XWindowChanges changes;
   XWindowAttributes wa;
 
   if (!XQueryTree(dpy, Scr.Root, &junk, &junk, &tops, &num))
     return;
-  topwin = 0;
-  topwin = Scr.FvwmRoot.stack_next->frame;
-  found = False;
 
   /********************************************************************
    * Locate the highest override_redirect window above our target, and
    * the highest of our windows below it.
    ********************************************************************/
-  for (i = 0; i < num; i++)  {
-    if (tops[i] == t->frame)  {
-      found = True;
+  for (i = 0; i < num && tops[i] != t->frame; i++)
+  {
+    /* look for target window in list */
+  }
+  for (; i < num; i++)
+  {
+    /* It might be just as well (and quicker) just to check for the absence of
+     * an FvwmContext instead of for override_redirect... */
+    if (!XGetWindowAttributes(dpy, tops[i], &wa))
+    {
+      continue;
     }
-    if (found)  {
-      /*
-	It might be just as well (and quicker) just to check for the absence of
-	an FvwmContext instead of for override_redirect...
-      */
-      if (XGetWindowAttributes(dpy, tops[i], &wa))  {
-	if (wa.override_redirect == True)  {
-	  /*  There's always at least 1 OR lurking somewhere...NoFocusWin.  */
-	  if (i > 0)  {
-	    if (XFindContext(dpy, tops[i - 1],
-			     FvwmContext, (caddr_t *) &junkwin) != XCNOENT)  {
-	      /*  save next lower non-OR win below highest OR...  */
-	      topwin = tops[i - 1];
-            }
-          }
-          OR_Above = tops[i];
-        }
-      }
-      /*
-	else  {
-	DBUG("raise_over_unmanaged",
-	"Error -  XGetWindowAttributes failed for window %8.8lx!\n",
-	tops[i]);
-	}
-      */
-    }    /*  end  if found  */
-  }      /*  end  for  */
+    if (wa.override_redirect == True && wa.class != InputOnly &&
+	tops[i] != Scr.NoFocusWin)
+    {
+      OR_Above = tops[i];
+    }
+  } /* end for */
 
   /********************************************************************
    * Count the windows we need to restack, then build the stack list.
    ********************************************************************/
   if (OR_Above)
   {
-    i = 0;
-    count = 0;
-    found = False;
-    for (t2 = t; (t2 != &Scr.FvwmRoot && ! found); t2 = t2->stack_prev)  {
+    for (count = 0, t2 = Scr.FvwmRoot.stack_next; t2 != &Scr.FvwmRoot;
+	 t2 = t2->stack_next)
+    {
       count++;
-      if (IS_ICONIFIED(t2) && ! IS_ICON_SUPPRESSED(t2))  {
-	if (t2->icon_w != None)  {
+      if (IS_ICONIFIED(t2) && ! IS_ICON_SUPPRESSED(t2))
+      {
+	if (t2->icon_w != None)
+	{
 	  count++;
         }
-	if (t2->icon_pixmap_w != None)  {
+	if (t2->icon_pixmap_w != None)
+	{
 	  count++;
         }
       }
-      if (t2->frame == topwin)  {
-	found = True;
+      if (t2 == t)
+      {
+	break;
       }
     }
 
-    i = 0;
-    found = False;
-    if (count > 0)  {
-      topwin = Scr.FvwmRoot.stack_next->frame;
+    if (count > 0)
+    {
       wins = (Window*) safemalloc (count * sizeof (Window));
-      for (t2 = t; t2 != &Scr.FvwmRoot && ! found; t2 = t2->stack_prev)  {
+      for (i = 0, t2 = Scr.FvwmRoot.stack_next; t2 != &Scr.FvwmRoot;
+	   t2 = t2->stack_next)
+      {
 	wins[i++] = t2->frame;
-	if (IS_ICONIFIED(t2) && ! IS_ICON_SUPPRESSED(t2))  {
-	  if (t2->icon_w != None)  {
+	if (IS_ICONIFIED(t2) && ! IS_ICON_SUPPRESSED(t2))
+	{
+	  if (t2->icon_w != None)
+	  {
 	    wins[i++] = t2->icon_w;
           }
-	  if (t2->icon_pixmap_w != None)  {
+	  if (t2->icon_pixmap_w != None)
+	  {
 	    wins[i++] = t2->icon_pixmap_w;
           }
         }
-	if (t2->frame == topwin)  {
-	  found = True;
-        }
+	if (t2 == t)
+	{
+	  break;
+	}
       }
 
       memset(&changes, '\0', sizeof(changes));
@@ -200,12 +374,11 @@ static void raise_over_unmanaged(FvwmWindow *t)
       changes.stack_mode = Above;
       flags = CWSibling|CWStackMode;
 
-      XConfigureWindow (dpy, topwin, flags, &changes);
+      XConfigureWindow (dpy, t->frame/*topwin*/, flags, &changes);
       XRestackWindows (dpy, wins, i);
       free (wins);
     }
-  }       /*  end - we found an OR above our target  */
-
+  }/*  end - we found an OR above our target  */
 
   XFree (tops);
   return;
@@ -217,6 +390,10 @@ static Bool must_move_transients(
 {
   *found_transient = False;
 
+  if (IS_ICONIFIED(t))
+  {
+    return False;
+  }
   /* raise */
   if ((!do_lower && DO_RAISE_TRANSIENT(t)) ||
       (do_lower && DO_LOWER_TRANSIENT(t)))
@@ -266,14 +443,82 @@ static Bool must_move_transients(
   return False;
 }
 
+static void restack_windows(
+  FvwmWindow *r, FvwmWindow *s, int count, Bool do_broadcast_all)
+{
+  FvwmWindow *t;
+  unsigned int flags;
+  int i;
+  XWindowChanges changes;
+  Window *wins;
+
+  if (count <= 0)
+  {
+    FvwmWindow *prev;
+
+    for (t = r, prev = NULL; prev != s; prev = t, t = t->stack_next)
+    {
+      count++;
+      if (IS_ICONIFIED(t) && !IS_ICON_SUPPRESSED(t))
+      {
+	count += 2;
+      }
+    }
+  }
+  /* restack the windows between r and s */
+  wins = (Window*) safemalloc (count * sizeof (Window));
+  i = 0;
+  for (t = r->stack_next; t != s; t = t->stack_next)
+  {
+    if (i >= count)
+    {
+      fvwm_msg (ERR, "restack_windows", "more transients than expected");
+      break;
+    }
+    wins[i++] = t->frame;
+    if (IS_ICONIFIED(t) && !IS_ICON_SUPPRESSED(t))
+    {
+      if(t->icon_w != None)
+	wins[i++] = t->icon_w;
+      if(t->icon_pixmap_w != None)
+	wins[i++] = t->icon_pixmap_w;
+    }
+  }
+
+  changes.sibling = s->frame;
+  if (changes.sibling != None)
+  {
+    changes.stack_mode = Above;
+    flags = CWSibling|CWStackMode;
+  }
+  else
+  {
+    changes.stack_mode = Below;
+    flags = CWStackMode;
+  }
+  XConfigureWindow (dpy, r->stack_next->frame, flags, &changes);
+  XRestackWindows (dpy, wins, count);
+  free(wins);
+  if (do_broadcast_all)
+  {
+    /* send out M_RESTACK for all windows, to make sure we don't forget
+     * anything. */
+    BroadcastRestackAllWindows();
+  }
+  else
+  {
+    /* send out (one or more) M_RESTACK packets for windows between r and s */
+    BroadcastRestack(r, s);
+  }
+
+  return;
+}
+
 static void RaiseOrLowerWindow(
   FvwmWindow *t, Bool do_lower, Bool allow_recursion, Bool is_new_window)
 {
   FvwmWindow *s, *r, *t2, *next, tmp_r;
-  unsigned int flags;
-  int i, count;
-  XWindowChanges changes;
-  Window *wins;
+  int count;
   Bool do_move_transients;
   Bool found_transient;
   Bool no_movement;
@@ -289,7 +534,7 @@ static void RaiseOrLowerWindow(
    * recursion stuff for new windows because the must_move_transients() call
    * needs a properly ordered stack ring - but the new window is still at the
    * front of the stack ring. */
-  if (allow_recursion && !is_new_window)
+  if (allow_recursion && !is_new_window && !IS_ICONIFIED(t))
   {
     /*
      * This part makes Raise/Lower on a Transient act on its Main and sibling
@@ -309,6 +554,10 @@ static void RaiseOrLowerWindow(
       {
         if (t2->w == t->transientfor)
         {
+	  if (IS_ICONIFIED(t2))
+	  {
+	    break;
+	  }
           if ((!do_lower && DO_RAISE_TRANSIENT(t2)) ||
               (do_lower && DO_LOWER_TRANSIENT(t2))  )
           {
@@ -361,7 +610,7 @@ static void RaiseOrLowerWindow(
       {
 	next = t2->stack_next;
 	if ((IS_TRANSIENT(t2)) && (t2->transientfor == t->w) &&
-	    (t2->layer == t->layer))
+	    (t2->layer == t->layer) && !IS_ICONIFIED(t2))
 	{
 	  /* t2 is a transient to lower */
 	  count++;
@@ -372,12 +621,8 @@ static void RaiseOrLowerWindow(
 
 	  /* unplug it */
 	  remove_window_from_stack_ring(t2);
-
 	  /* put it above tmp_r */
-	  t2->stack_next = &tmp_r;
-	  t2->stack_prev = tmp_r.stack_prev;
-	    t2->stack_prev->stack_next = t2;
-	    tmp_r.stack_prev = t2;
+	  add_window_to_stack_ring_after(t2, tmp_r.stack_prev);
 	}
       }
       if (tmp_r.stack_next == &tmp_r)
@@ -403,11 +648,8 @@ static void RaiseOrLowerWindow(
 
     if (do_move_transients && tmp_r.stack_next != &tmp_r)
     {
-	/* insert all transients between r and s. */
-	r->stack_next = tmp_r.stack_next;
-	tmp_r.stack_next->stack_prev = r;
-	s->stack_prev = tmp_r.stack_prev;
-	tmp_r.stack_prev->stack_next = s;
+      /* insert all transients between r and s. */
+      add_windowlist_to_stack_ring_after(&tmp_r, r);
     }
 
     /*
@@ -415,7 +657,8 @@ static void RaiseOrLowerWindow(
     */
     add_window_to_stack_ring_after(t, s->stack_prev);
 
-    if (is_new_window && IS_TRANSIENT(t) && DO_STACK_TRANSIENT_PARENT(t))
+    if (is_new_window && IS_TRANSIENT(t) && DO_STACK_TRANSIENT_PARENT(t) &&
+	!IS_ICONIFIED(t))
     {
       /* now that the new transient is properly positioned in the stack ring,
        * raise/lower it again so that its parent is raised/lowered too */
@@ -427,49 +670,9 @@ static void RaiseOrLowerWindow(
     }
     else
     {
-      wins = (Window*) safemalloc (count * sizeof (Window));
-      i = 0;
-      for (t2 = r->stack_next; t2 != s; t2 = t2->stack_next)
-      {
-	if (i >= count)
-	{
-	  fvwm_msg (ERR, "RaiseOrLowerWindow", "more transients than expected");
-	  break;
-	}
-	wins[i++] = t2->frame;
-	if (IS_ICONIFIED(t2) && !IS_ICON_SUPPRESSED(t2))
-	{
-	  if(t2->icon_w != None)
-	    wins[i++] = t2->icon_w;
-	  if(t2->icon_pixmap_w != None)
-	    wins[i++] = t2->icon_pixmap_w;
-	}
-      }
-
-      changes.sibling = s->frame;
-      if (changes.sibling != None)
-      {
-	changes.stack_mode = Above;
-	flags = CWSibling|CWStackMode;
-      }
-      else
-      {
-	changes.stack_mode = Below;
-	flags = CWStackMode;
-      }
-
-      XConfigureWindow (dpy, r->stack_next->frame, flags, &changes);
-      XRestackWindows (dpy, wins, count);
-
-      /* send out (one or more) M_RESTACK packets for windows between r and s */
-      BroadcastRestack (r, s);
-      if (do_move_transients && tmp_r.stack_next != &tmp_r)
-      {
-	/* send out M_RESTACK for all windows, to make sure we don't forget
-	 * anything. */
-	BroadcastRestack (Scr.FvwmRoot.stack_next, Scr.FvwmRoot.stack_prev);
-      }
-      free (wins);
+      /* restack the windows between r and s */
+      restack_windows(
+	r, s, count, (do_move_transients && tmp_r.stack_next != &tmp_r));
     }
   }
 
@@ -477,7 +680,7 @@ static void RaiseOrLowerWindow(
   {
     /* This hack raises the target and all higher FVWM windows over any style
      * grabfocusoff override_redirect windows that may be above it. This is
-     * used to cope with ill-bahaved applications that insist on using
+     * used to cope with ill-behaved applications that insist on using
      * long-lived override_redirects. */
     if (Scr.bo.RaiseOverUnmanaged)
     {
@@ -524,7 +727,10 @@ static void RaiseOrLowerWindow(
   /* If the window has been raised, make sure the decorations are updated
    * immediately in case we are in a complex function (e.g. raise, unshade). */
   if (!do_lower)
-    DrawDecorations(t, DRAW_ALL, (Scr.Hilite == t), True, None);
+  {
+    XSync(dpy, 0);
+    handle_all_expose();
+  }
 }
 
 /*
@@ -535,13 +741,23 @@ static void RaiseOrLowerWindow(
 */
 void RaiseWindow(FvwmWindow *t)
 {
+  BroadcastPacket(M_RAISE_WINDOW, 3, t->w, t->frame, (unsigned long)t);
   RaiseOrLowerWindow(t, False, True, False);
+  focus_grab_buttons_on_pointer_window();
+#ifdef DEBUG_STACK_RING
+  verify_stack_ring_consistency();
+#endif
   return;
 }
 
 void LowerWindow(FvwmWindow *t)
 {
+  BroadcastPacket(M_LOWER_WINDOW, 3, t->w, t->frame, (unsigned long)t);
   RaiseOrLowerWindow(t, True, True, False);
+  focus_grab_buttons_on_pointer_window();
+#ifdef DEBUG_STACK_RING
+  verify_stack_ring_consistency();
+#endif
   return;
 }
 
@@ -645,6 +861,9 @@ HandleUnusualStackmodes(unsigned int stack_mode, FvwmWindow *r, Window rw,
     break;
   }
 /*  DBUG("HandleUnusualStackmodes", "\t---> %d\n", restack);*/
+#ifdef DEBUG_STACK_RING
+  verify_stack_ring_consistency();
+#endif
   return restack;
 }
 
@@ -656,6 +875,7 @@ HandleUnusualStackmodes(unsigned int stack_mode, FvwmWindow *r, Window rw,
     libfvwm eventually, along with some other chain manipulation functions.
 */
 
+#if 0
 /*
     ResyncFvwmStackRing -
     Rebuilds the stacking order ring of FVWM-managed windows. For use in cases
@@ -707,6 +927,12 @@ static void ResyncFvwmStackRing (void)
       /* Pluck from chain. */
       remove_window_from_stack_ring(t1);
       add_window_to_stack_ring_after(t1, t2->stack_prev);
+      if (t2 != &Scr.FvwmRoot && t2->layer > t1->layer)
+      {
+	/* oops, now our stack ring is out of order! */
+	/* emergency fix */
+	t1->layer = t2->layer;
+      }
       t2 = t1;
     }
   }
@@ -715,6 +941,7 @@ static void ResyncFvwmStackRing (void)
 
   XFree (children);
 }
+#endif
 
 /* same as above but synchronizes the stacking order in X from the stack ring.
  */
@@ -725,7 +952,7 @@ static void ResyncXStackingOrder(void)
   int count;
   int i;
 
-  for (count = 0, t = Scr.FvwmRoot.next; t != None; count++, t = t->next)
+  for (count = 0, t = Scr.FvwmRoot.next; t != NULL; count++, t = t->next)
     ;
   if (count > 0)
   {
@@ -746,24 +973,31 @@ static void ResyncXStackingOrder(void)
     free(wins);
     /* send out M_RESTACK for all windows, to make sure we don't forget
      * anything. */
-    BroadcastRestack(Scr.FvwmRoot.stack_next, Scr.FvwmRoot.stack_prev);
+    BroadcastRestackAllWindows();
   }
 }
 
 
 /* send RESTACK packets for all windows between s1 and s2 */
-void BroadcastRestack (FvwmWindow *s1, FvwmWindow *s2)
+static void BroadcastRestack(FvwmWindow *s1, FvwmWindow *s2)
 {
   FvwmWindow *t, *t2;
   int num, i;
   unsigned long *body, *bp, length;
   extern Time lastTimestamp;
 
+  if (s2 == &Scr.FvwmRoot)
+  {
+    s2 = s2->stack_prev;
+    if (s2 == &Scr.FvwmRoot)
+      return;
+  }
   if (s1 == &Scr.FvwmRoot)
   {
     t = s1->stack_next;
+    if (t == &Scr.FvwmRoot)
+      return;
     /* t has been moved to the top of stack */
-
     BroadcastPacket (M_RAISE_WINDOW, 3, t->w, t->frame, (unsigned long)t);
     if (t->stack_next == s2)
     {
@@ -775,9 +1009,14 @@ void BroadcastRestack (FvwmWindow *s1, FvwmWindow *s2)
   {
     t = s1;
   }
-  for (t2 = t, num = 1 ; t2 != s2; t2 = t2->stack_next, num++)
+  if (s1 == s2)
+  {
+    /* A useful M_RESTACK packet must contain at least two windows. */
+    return;
+  }
+  for (t2 = t, num = 1 ; t2 != s2 && t != &Scr.FvwmRoot;
+       t2 = t2->stack_next, num++)
     ;
-
   length = FvwmPacketHeaderSize + 3*num;
   body = (unsigned long *) safemalloc (length*sizeof(unsigned long));
 
@@ -786,16 +1025,26 @@ void BroadcastRestack (FvwmWindow *s1, FvwmWindow *s2)
   *(bp++) = M_RESTACK;
   *(bp++) = length;
   *(bp++) = lastTimestamp;
-  for (t2 = t; t2 != s2; t2 = t2->stack_next)
-   {
-     *(bp++) = t2->w;
-     *(bp++) = t2->frame;
-     *(bp++) = (unsigned long)t2;
-   }
-   for (i = 0; i < npipes; i++)
-     PositiveWrite(i, body, length*sizeof(unsigned long));
+  for (t2 = t; num != 0; num--, t2 = t2->stack_next)
+  {
+    *(bp++) = t2->w;
+    *(bp++) = t2->frame;
+    *(bp++) = (unsigned long)t2;
+  }
+  for (i = 0; i < npipes; i++)
+    PositiveWrite(i, body, length*sizeof(unsigned long));
+  free(body);
+#ifdef DEBUG_STACK_RING
+  verify_stack_ring_consistency();
+#endif
 
-   free (body);
+  return;
+}
+
+void BroadcastRestackAllWindows(void)
+{
+  BroadcastRestack(Scr.FvwmRoot.stack_next, Scr.FvwmRoot.stack_prev);
+  return;
 }
 
 /* send RESTACK packets for t, t->stack_prev and t->stack_next */
@@ -831,52 +1080,209 @@ int get_layer(FvwmWindow *t)
   return t->layer;
 }
 
-void
-new_layer (FvwmWindow *tmp_win, int layer)
+/* This function recursively finds the transients of the window t and sets their
+ * is_in_transient_subtree flag.  If a layer is given, only windows in this
+ * layer are checked.  If the layer is < 0, all windows are considered.
+ */
+void mark_transient_subtree(
+  FvwmWindow *t, int layer, int mark_mode, Bool do_ignore_icons,
+  Bool use_window_group_hint)
 {
-  FvwmWindow *t2, *next;
+  FvwmWindow *s;
+  FvwmWindow *start;
+  FvwmWindow *end;
+  Bool is_finished;
 
+  if (layer >= 0 && t->layer != layer)
+    return;
+  /* find out on which windows to operate */
+  if (layer >= 0)
+  {
+    /* only work on the given layer */
+    start = &Scr.FvwmRoot;
+    end = &Scr.FvwmRoot;
+    for (s = Scr.FvwmRoot.stack_next; s != &Scr.FvwmRoot && s->layer >= layer;
+	 s = s->stack_next)
+    {
+      if (s->layer == layer)
+      {
+	if (start == &Scr.FvwmRoot)
+	  start = s;
+	end = s->stack_next;
+      }
+    }
+  }
+  else
+  {
+    /* work on complete window list */
+    start = Scr.FvwmRoot.stack_next;
+    end = &Scr.FvwmRoot;
+  }
+  /* clean the temporary flag in all windows and precalculate the transient
+   * frame windows */
+  for (s = Scr.FvwmRoot.stack_next; s != &Scr.FvwmRoot; s = s->stack_next)
+  {
+    SET_IN_TRANSIENT_SUBTREE(s, 0);
+    if (IS_TRANSIENT(s) && (layer < 0 || layer == s->layer))
+    {
+      s->pscratch = get_transientfor_fvwmwindow(s);
+    }
+    else
+    {
+      s->pscratch = NULL;
+    }
+  }
+  /* now loop over the windows and mark the ones we need to move */
+  SET_IN_TRANSIENT_SUBTREE(t, 1);
+  is_finished = False;
+  while (!is_finished)
+  {
+    FvwmWindow *r;
+    FvwmWindow *u;
+
+    /* recursively search for all transient windows */
+    is_finished = True;
+    for (s = start; s != end; s = s->stack_next)
+    {
+      Bool use_group_hint = False;
+
+      if (IS_IN_TRANSIENT_SUBTREE(s))
+	continue;
+      if (DO_ICONIFY_WINDOW_GROUPS(s) && s->wmhints &&
+	  (s->wmhints->flags & WindowGroupHint) &&
+	  (s->wmhints->window_group != None) &&
+	  (s->wmhints->window_group != s->w) &&
+	  (s->wmhints->window_group != Scr.Root))
+      {
+	use_group_hint = True;
+      }
+      if (!IS_TRANSIENT(s) && !use_group_hint)
+	continue;
+      r = (FvwmWindow *)s->pscratch;
+      if (do_ignore_icons && r && IS_ICONIFIED(r))
+	continue;
+      if (IS_TRANSIENT(s))
+      {
+	if (r && IS_IN_TRANSIENT_SUBTREE(r) &&
+	    ((mark_mode == MARK_ALL) ||
+	     (mark_mode == MARK_LOWER && DO_LOWER_TRANSIENT(r)) ||
+	     (mark_mode == MARK_RAISE && DO_RAISE_TRANSIENT(r))))
+	{
+	  /* have to move this one too */
+	  SET_IN_TRANSIENT_SUBTREE(s, 1);
+	  /* need another scan through the list */
+	  is_finished = False;
+	  continue;
+	}
+      }
+      if (use_group_hint)
+      {
+	for (u = start; u != end; u = u->stack_next)
+	{
+	  if (u->w == s->wmhints->window_group ||
+	      (u->wmhints && (u->wmhints->flags & WindowGroupHint) &&
+	       u->wmhints->window_group == s->wmhints->window_group))
+	  {
+	    if (IS_IN_TRANSIENT_SUBTREE(u))
+	    {
+	      /* have to move this one too */
+	      SET_IN_TRANSIENT_SUBTREE(s, 1);
+	      /* need another scan through the list */
+	      is_finished = False;
+	    }
+	  }
+	}
+      }
+    } /* for */
+  } /* while */
+
+  return;
+}
+
+static int collect_transients_recursive(
+  FvwmWindow *t, FvwmWindow *list_head, int layer, Bool do_lower)
+{
+  FvwmWindow *s;
+  int count = 0;
+
+  mark_transient_subtree(
+    t, layer, (do_lower) ? MARK_LOWER : MARK_RAISE, True, False);
+  /* now collect the marked windows in a separate list */
+  for (s = Scr.FvwmRoot.stack_next; s != &Scr.FvwmRoot; )
+  {
+    FvwmWindow *tmp;
+
+    tmp = s->stack_next;
+    if (IS_IN_TRANSIENT_SUBTREE(s))
+    {
+      remove_window_from_stack_ring(s);
+      add_window_to_stack_ring_after(s, list_head->stack_prev);
+      count++;
+      if (IS_ICONIFIED(t) && !IS_ICON_SUPPRESSED(t))
+	count += 2;
+    }
+    s = tmp;
+  }
+
+  return count;
+}
+
+void new_layer(FvwmWindow *tmp_win, int layer)
+{
+  FvwmWindow *s;
+  FvwmWindow *target;
+  FvwmWindow *prev;
+  FvwmWindow list_head;
+  int add_after_layer;
+  int count;
+
+  tmp_win = get_transientfor_top_fvwmwindow(tmp_win);
+  if (layer == tmp_win->layer)
+    return;
+  list_head.stack_next = &list_head;
+  list_head.stack_prev = &list_head;
+  count = collect_transients_recursive(
+    tmp_win, &list_head, tmp_win->layer, (layer < tmp_win->layer));
+  if (count == 0)
+  {
+    /* no windows to move */
+    return;
+  }
+
+  add_after_layer = layer;
   if (layer < tmp_win->layer)
   {
-    tmp_win->layer = layer;
-    /* temporary fix; new transient handling code assumes the layers are
-     * properly arranged when RaiseOrLowerWindow() is called. */
-    ResyncFvwmStackRing();
-    /* domivogt (9-Sep-1999): this was RaiseWindow before. The intent may
-     * have been good (put the window on top of the new lower layer), but it
-     * doesn't work with applications like the gnome panel that use the layer
-     * hint as a poor man's raise/lower, i.e. if panel is on the top level
-     * and requests to be lowered to the normal level it stays on top of all
-     * normal windows. This is not what was intended by lowering. */
-    LowerWindow(tmp_win);
+    /* lower below the windows in the new (lower) layer */
+    add_after_layer = layer;
   }
-  else if (layer > tmp_win->layer)
+  else
+  {
+    /* raise above the windows in the new (higher) layer */
+    add_after_layer = layer + 1;
+  }
+  /* find the place to insert the windows */
+  for (target = Scr.FvwmRoot.stack_next; target != &Scr.FvwmRoot;
+       target = target->stack_next)
+  {
+    if (target->layer < add_after_layer)
     {
-      if (DO_RAISE_TRANSIENT(tmp_win))
-	{
-	  /* this could be done much more efficiently */
-	  for (t2 = Scr.FvwmRoot.stack_next; t2 != &Scr.FvwmRoot; t2 = next)
-	    {
-	      next = t2->stack_next;
-	      if ((IS_TRANSIENT(t2)) &&
-		  (t2->transientfor == tmp_win->w) &&
-		  (t2 != tmp_win) &&
-		  (t2->layer >= tmp_win->layer) &&
-		  (t2->layer < layer))
-		{
-		  t2->layer = layer;
-		  LowerWindow(t2);
-		}
-	    }
-	}
-      tmp_win->layer = layer;
-      /* see comment above */
-      ResyncFvwmStackRing();
-      /* see comment above */
-      RaiseWindow(tmp_win);
+      /* add all windows before the current window */
+      break;
     }
+  }
+  /* insert windows at new position */
+  add_windowlist_to_stack_ring_after(&list_head, target->stack_prev);
+  prev = NULL;
+  for (s = list_head.stack_next; prev != list_head.stack_prev;
+       prev = s, s = s->stack_next)
+  {
+    s->layer = layer;
+    GNOME_SetLayer(tmp_win);
+  }
+  /* move the windows without modifying their stacking order */
+  restack_windows(list_head.stack_next->stack_prev, target, count, (count > 1));
 
-  GNOME_SetLayer (tmp_win);
+  return;
 }
 
 /* ----------------------------- common functions -------------------------- */
@@ -894,7 +1300,8 @@ void init_stack_and_layers(void)
   return;
 }
 
-Bool is_on_top_of_layer(FvwmWindow *fw)
+
+static Bool is_on_top_of_layer_ignore_rom(FvwmWindow *fw)
 {
   FvwmWindow *t;
   Bool ontop = True;
@@ -905,7 +1312,14 @@ Bool is_on_top_of_layer(FvwmWindow *fw)
     {
       break;
     }
-    if (overlap(fw, t))
+    if (t->Desk != fw->Desk)
+    {
+      continue;
+    }
+    /* For RaiseOverUnmanaged we can not determine if the window is on top by
+     * checking if the window overlaps another one.  If it was below unmanaged
+     * windows, but on top of its layer, it would be considered on top. */
+    if (Scr.bo.RaiseOverUnmanaged || overlap(fw, t))
     {
       if (!IS_TRANSIENT(t) || t->transientfor != fw->w ||
 	  !DO_RAISE_TRANSIENT(fw))
@@ -919,9 +1333,18 @@ Bool is_on_top_of_layer(FvwmWindow *fw)
   return ontop;
 }
 
+Bool is_on_top_of_layer(FvwmWindow *fw)
+{
+  if (Scr.bo.RaiseOverUnmanaged)
+  {
+    return False;
+  }
+  return is_on_top_of_layer_ignore_rom(fw);
+}
+
 /* ----------------------------- built in functions ------------------------ */
 
-void raise_function(F_CMD_ARGS)
+void CMD_Raise(F_CMD_ARGS)
 {
   if (DeferExecution(eventp,&w,&tmp_win,&context, CRS_SELECT,ButtonRelease))
     return;
@@ -929,7 +1352,7 @@ void raise_function(F_CMD_ARGS)
   RaiseWindow(tmp_win);
 }
 
-void lower_function(F_CMD_ARGS)
+void CMD_Lower(F_CMD_ARGS)
 {
   if (DeferExecution(eventp,&w,&tmp_win,&context, CRS_SELECT, ButtonRelease))
     return;
@@ -937,22 +1360,27 @@ void lower_function(F_CMD_ARGS)
   LowerWindow(tmp_win);
 }
 
-void raiselower_func(F_CMD_ARGS)
+void CMD_RaiseLower(F_CMD_ARGS)
 {
   Bool ontop;
 
   if (DeferExecution(eventp,&w,&tmp_win,&context, CRS_SELECT,ButtonRelease))
+  {
+#ifdef DEBUG_STACK_RING
+    dump_stack_ring();
+#endif
     return;
-
-  ontop = is_on_top_of_layer(tmp_win);
-
+  }
+  ontop = is_on_top_of_layer_ignore_rom(tmp_win);
   if (ontop)
     LowerWindow(tmp_win);
   else
     RaiseWindow(tmp_win);
+
+  return;
 }
 
-void change_layer(F_CMD_ARGS)
+void CMD_Layer(F_CMD_ARGS)
 {
   int n, layer, val[2];
   char *token;
@@ -965,38 +1393,40 @@ void change_layer(F_CMD_ARGS)
 
   token = PeekToken(action, NULL);
   if (StrEquals("default", token))
+  {
+    layer = tmp_win->default_layer;
+  }
+  else
+  {
+    n = GetIntegerArguments(action, NULL, val, 2);
+
+    layer = tmp_win->layer;
+    if ((n == 1) ||
+	((n == 2) && (val[0] != 0)))
+    {
+      layer += val[0];
+    }
+    else if ((n == 2) && (val[1] >= 0))
+    {
+      layer = val[1];
+    }
+    else
     {
       layer = tmp_win->default_layer;
     }
-  else
-    {
-      n = GetIntegerArguments(action, NULL, val, 2);
-
-      layer = tmp_win->layer;
-      if ((n == 1) ||
-	  ((n == 2) && (val[0] != 0)))
-	{
-	  layer += val[0];
-	}
-      else if ((n == 2) && (val[1] >= 0))
-	{
-	  layer = val[1];
-	}
-      else
-	{
-	  layer = tmp_win->default_layer;
-	}
-    }
-
+  }
   if (layer < 0)
-    {
-      layer = 0;
-    }
+  {
+    layer = 0;
+  }
+  new_layer(tmp_win, layer);
 
-  new_layer (tmp_win, layer);
+#ifdef DEBUG_STACK_RING
+  verify_stack_ring_consistency();
+#endif
 }
 
-void SetDefaultLayers(F_CMD_ARGS)
+void CMD_DefaultLayers(F_CMD_ARGS)
 {
   char *bot = NULL;
   char *def = NULL;
@@ -1005,43 +1435,47 @@ void SetDefaultLayers(F_CMD_ARGS)
 
   bot = PeekToken(action, &action);
   if (bot)
+  {
+    i = atoi (bot);
+    if (i < 0)
     {
-       i = atoi (bot);
-       if (i < 0)
-         {
-           fvwm_msg(ERR,"DefaultLayers", "Layer must be non-negative." );
-         }
-       else
-         {
-           Scr.BottomLayer = i;
-         }
+      fvwm_msg(ERR,"DefaultLayers", "Layer must be non-negative." );
     }
+    else
+    {
+      Scr.BottomLayer = i;
+    }
+  }
 
   def = PeekToken(action, &action);
   if (def)
+  {
+    i = atoi (def);
+    if (i < 0)
     {
-       i = atoi (def);
-       if (i < 0)
-         {
-           fvwm_msg(ERR,"DefaultLayers", "Layer must be non-negative." );
-         }
-       else
-         {
-           Scr.DefaultLayer = i;
-         }
+      fvwm_msg(ERR,"DefaultLayers", "Layer must be non-negative." );
     }
+    else
+    {
+      Scr.DefaultLayer = i;
+    }
+  }
 
   top = PeekToken(action, &action);
   if (top)
+  {
+    i = atoi (top);
+    if (i < 0)
     {
-       i = atoi (top);
-       if (i < 0)
-         {
-           fvwm_msg(ERR,"DefaultLayers", "Layer must be non-negative." );
-         }
-       else
-         {
-           Scr.TopLayer = i;
-         }
+      fvwm_msg(ERR,"DefaultLayers", "Layer must be non-negative." );
     }
+    else
+    {
+      Scr.TopLayer = i;
+    }
+  }
+
+#ifdef DEBUG_STACK_RING
+  verify_stack_ring_consistency();
+#endif
 }
