@@ -4,6 +4,10 @@
  *  Copyright 1994,  Mike Finger (mfinger@mermaid.micro.umn.edu or
  *                               Mike_Finger@atk.com)
  *
+ * Two little hacks by DYB: 1) make size of a strip in hidden state tunable
+ * (with *FvwmTaskBar: AutoHide <pixels>); 2) make process of TaskBar
+ * hiding/unhiding more smooth.
+ *
  * Minor hack by TKP to enable autohide to work with pages (leaves 2 pixels
  * visible so that the 1 pixel panframes don't get in the way)
  *
@@ -66,6 +70,8 @@
 
 #include "libs/Module.h"
 #include "libs/fvwmlib.h"  /* for pixmaps routines */
+#include "libs/FScreen.h"
+#include "libs/FShape.h"
 #include "libs/safemalloc.h"
 #include "libs/fvwmsignal.h"
 #include "libs/Colorset.h"
@@ -93,6 +99,8 @@
 unsigned char gray_bits[] = { 0xaa, 0x55, 0xaa, 0x55,
                               0xaa, 0x55, 0xaa, 0x55 };
 
+#define DEFAULT_VISIBLE_PIXELS 3
+
 /* File type information */
 int   Fvwm_fd[2];
 int   x_fd;
@@ -192,11 +200,18 @@ int UseSkipList    = False,
     AutoFocus      = False,
     HighlightFocus = False,
     DeskOnly       = False,
+    PageOnly       = False,
+    ScreenOnly     = False,
     NoBrightFocus  = False,
     ThreeDfvwm     = False,
     RowsNumber     = 1;
 
-unsigned int ScreenWidth, ScreenHeight;
+int VisiblePixels  = DEFAULT_VISIBLE_PIXELS;
+/* This macro should be used when calculating geometry */
+#define VISIBLE_PIXELS() (VisiblePixels < win_height? VisiblePixels:win_height)
+
+rectangle screen_g;
+rectangle global_scr_g;
 
 int NRows, RowHeight, Midline;
 
@@ -213,7 +228,9 @@ extern TipStruct Tip;
 extern int StartButtonWidth, StartButtonHeight;
 extern char *StartPopup;
 
-char *ImagePath   = NULL;
+char *ImagePath = NULL;
+char *XineramaConfig = NULL;
+static int fscreen = 0;
 
 static void ParseConfig( void );
 static void ParseConfigLine(char *tline);
@@ -249,22 +266,32 @@ int main(int argc, char **argv)
   if (s != NULL)
     temp = s + 1;
 
-  /* Setup my name */
-  Module = safemalloc(strlen(temp)+2);
-  strcpy(Module,"*");
-  strcat(Module, temp);
-  Clength = strlen(Module);
-
 #ifdef I18N_MB
   setlocale(LC_CTYPE, "");
 #endif
 
   if((argc != 6)&&(argc != 7)) {
-    fprintf(stderr,"%s Version %s should only be executed by fvwm!\n",Module,
+    fprintf(stderr,"%s Version %s should only be executed by fvwm!\n",temp,
 	    VERSION);
     exit(1);
   }
 
+  /* alise support */
+  if (argc == 7 && argv[6] != NULL)
+  {
+    Module = safemalloc(strlen(argv[6])+2);
+    strcpy(Module,"*");
+    strcat(Module, argv[6]);
+    Clength = strlen(Module);
+  }
+  else
+  {
+    Module = safemalloc(strlen(temp)+2);
+    strcpy(Module,"*");
+    strcat(Module, temp);
+    Clength = strlen(Module);
+  }
+
   /* setup fvwm pipes */
   Fvwm_fd[0] = atoi(argv[1]);
   Fvwm_fd[1] = atoi(argv[2]);
@@ -427,6 +454,7 @@ void ProcessMessage(unsigned long type,unsigned long *body)
   Picture p;
   struct ConfigWinPacket  *cfgpacket;
   int iconified;
+  rectangle new_g;
 
 /*    memset(&p, 0, sizeof(Picture)); */
 
@@ -451,10 +479,14 @@ void ProcessMessage(unsigned long type,unsigned long *body)
   case M_ADD_WINDOW:
   case M_CONFIGURE_WINDOW:
     cfgpacket = (ConfigWinPacket *) body;
+    new_g.x = cfgpacket->frame_x;
+    new_g.y = cfgpacket->frame_y;
+    new_g.width = cfgpacket->frame_width;
+    new_g.height = cfgpacket->frame_height;
     if (!ShowTransients && (IS_TRANSIENT(cfgpacket)))
       break;
     /* Notes: users cannot modify the win_width of the taskbar (see
-       the hints in StartMeup; it is a good idea). So "we have to" modify
+       the hints in StartMeUp; it is a good idea). So "we have to" modify
        the win_width here by modifying the WMNormalHints. Moreover, 1. we want
        that the TaskBar width with its border is execately the screen width
        2. be carful with dynamic style change */
@@ -469,32 +501,44 @@ void ProcessMessage(unsigned long type,unsigned long *body)
 	long dumy;
 
 	win_border = (short)cfgpacket->border_width;
-	win_width = ScreenWidth-(win_border<<1);
+	win_width = screen_g.width-(win_border<<1);
 	win_title_height = (int)cfgpacket->title_height;
 	win_has_title = HAS_TITLE(cfgpacket);
 	win_has_bottom_title = HAS_BOTTOM_TITLE(cfgpacket);
 
 	if (AutoStick)
 	{
-	  win_x = win_border;
-	  if (!win_is_shaded) {
+	  win_x = screen_g.x + win_border;
+	  if (!win_is_shaded)
+	  {
 	    if (win_y > Midline)
-	      win_y = ScreenHeight -
-		(AutoHide ? 2 - win_title_height*win_has_bottom_title :
+	    {
+	      win_y = screen_g.height -
+		(AutoHide ?
+		 VISIBLE_PIXELS() - win_title_height*win_has_bottom_title :
 		 win_height + win_border);
+	    }
 	    else
-	      win_y = AutoHide ? 2 + win_title_height*win_has_bottom_title
+	    {
+	      win_y = AutoHide ?
+		VISIBLE_PIXELS() + win_title_height*win_has_bottom_title
 		- win_height : win_border + win_title_height;
+	    }
 	  }
 	  else
 	  {
-	   if (win_y > Midline)
-	    win_y = ScreenHeight - win_border -
-	      (win_has_bottom_title)*win_height;
-	  else
-	     win_y = win_title_height + win_border -
-	       (win_has_bottom_title)*win_height;
+	    if (win_y > Midline)
+	    {
+	      win_y = screen_g.height - win_border -
+		(win_has_bottom_title)*win_height;
+	    }
+	    else
+	    {
+	      win_y = win_title_height + win_border -
+		(win_has_bottom_title)*win_height;
+	    }
 	  }
+	  win_y += screen_g.y;
 	}
 
 	XGetWMNormalHints(dpy,win,&hints,&dumy);
@@ -518,25 +562,34 @@ void ProcessMessage(unsigned long type,unsigned long *body)
 	win_is_shaded = IS_SHADED(cfgpacket);
 	if (win_is_shaded)
 	{
-	   if (win_y > Midline)
-	    win_y = ScreenHeight - win_border -
+	  if (win_y > Midline)
+	  {
+	    win_y = screen_g.height - win_border -
 	      (win_has_bottom_title)*win_height;
+	  }
 	  else
-	     win_y = win_title_height + win_border -
-	       (win_has_bottom_title)*win_height;
+	  {
+	    win_y = win_title_height + win_border -
+	      (win_has_bottom_title)*win_height;
+	  }
 	}
 	else
 	{
 	  if (win_y > Midline)
-	    win_y = ScreenHeight - win_border - win_height;
+	  {
+	    win_y = screen_g.height - win_border - win_height;
+	  }
 	  else
+	  {
 	    win_y = win_title_height + win_border;
+	  }
 	  if (AutoHide)
 	  {
 	    WindowState = 1;
 	    SetAlarm(HIDE_TASK_BAR);
 	  }
 	}
+	win_y += screen_g.y;
 	XMoveWindow(dpy, win, win_x, win_y);
       }
       break;
@@ -547,44 +600,92 @@ void ProcessMessage(unsigned long type,unsigned long *body)
       int remove=0;
       int add = 0;
       int skiplist = 0;
+      rectangle *pold_g;
+      Bool is_on_desk = True;
+      Bool is_changing_desk = False;
+      Bool is_on_page = True;
+      Bool is_changing_page = False;
+      Bool is_on_screen = True;
+      Bool is_changing_screen = False;
+      Bool is_managed;
 
-      if (GetDeskNumber(&windows,i,&Desk) && DeskOnly)
+      if (!GetItemGeometry(&windows, i, &pold_g))
       {
-        if (DeskNumber != Desk && DeskNumber == cfgpacket->desk
-	    && !IsItemIndexSticky(&windows,i))
+	pold_g = &new_g;
+      }
+
+      if (DeskOnly && GetDeskNumber(&windows,i,&Desk))
+      {
+	if (IsItemIndexSticky(&windows,i))
 	{
-          /* window moving to current desktop */
-          add = 1;
-        }
-        if (DeskNumber != cfgpacket->desk && DeskNumber == Desk)
+	  is_on_desk = True;
+	  is_changing_desk = False;
+	}
+	else if (!IsItemIndexSticky(&windows,i))
 	{
-          /* window moving to another desktop */
-          remove = 1;
-        }
+	  is_on_desk = (DeskNumber != cfgpacket->desk);
+	  is_changing_desk = (Desk != cfgpacket->desk);
+	}
 	UpdateItemIndexDesk(&windows, i, cfgpacket->desk);
       }
-      if (UseSkipList && (!DeskOnly || DeskNumber == cfgpacket->desk)) {
+      if (PageOnly)
+      {
+	is_on_page = fvwmrect_do_rectangles_intersect(&new_g, &global_scr_g);
+	is_changing_page =
+	  (fvwmrect_do_rectangles_intersect(pold_g, &global_scr_g) !=
+	   is_on_page);
+      }
+      if (ScreenOnly)
+      {
+	is_on_screen = fvwmrect_do_rectangles_intersect(&new_g, &screen_g);
+	is_changing_screen =
+	  (fvwmrect_do_rectangles_intersect(pold_g, &screen_g) !=
+	   is_on_screen);
+      }
+      UpdateItemIndexGeometry(&windows, i, &new_g);
+
+      is_managed = (is_on_desk && is_on_page && is_on_screen);
+      if (is_changing_desk || is_changing_page || is_changing_screen)
+      {
+	if (is_managed)
+	{
+	  /* window entering managed area of the virtual desktop */
+	  add = 1;
+	}
+	else
+	{
+	  /* window leaving managed area of the virtual desktop */
+	  remove = 1;
+	}
+      }
+
+      if (UseSkipList && is_managed)
+      {
 	skiplist = IsItemIndexSkipWindowList(&windows,i);
-	if (DO_SKIP_WINDOW_LIST(cfgpacket) && !skiplist) {
+	if (DO_SKIP_WINDOW_LIST(cfgpacket) && !skiplist)
+	{
+	  add = 0;
 	  remove = 1;
 	}
 	if (!DO_SKIP_WINDOW_LIST(cfgpacket) && skiplist) {
 	  add = 1;
+	  remove = 0;
 	}
       }
 
       UpdateItemGSFRFlags(&windows, cfgpacket);
-      if (remove) {
+      if (remove)
+      {
 	RemoveButton(&buttons, i);
 	redraw = 1;
       }
-      else if (add) {
-	 AddButton(&buttons, ItemName(&windows,i),
-		   GetItemPicture(&windows,i), BUTTON_UP, i,
-		   IsItemIndexIconified(&windows, i));
-          redraw = 1;
+      else if (add)
+      {
+	AddButton(&buttons, ItemName(&windows,i),
+		  GetItemPicture(&windows,i), BUTTON_UP, i,
+		  IsItemIndexIconified(&windows, i));
+	redraw = 1;
       }
-
     }
     else
     {
@@ -593,6 +694,7 @@ void ProcessMessage(unsigned long type,unsigned long *body)
       if (Count > COUNT_LIMIT)
 	Count = 0;
       i = FindItem(&windows, cfgpacket->w);
+      UpdateItemIndexGeometry(&windows, i, &new_g);
     }
     if (i != -1)
     {
@@ -635,15 +737,23 @@ void ProcessMessage(unsigned long type,unsigned long *body)
       break;
     if (UpdateButton(&buttons, i, string, DONT_CARE) == -1)
     {
+      rectangle *pold_g;
+
       if (GetDeskNumber(&windows, i, &Desk) == 0)
 	break;
-      if ((!DeskOnly || Desk == DeskNumber) &&
-	  (!UseSkipList || !IsItemIndexSkipWindowList(&windows, i)))
-      {
-	AddButton(&buttons, string, NULL, BUTTON_UP, i,
-		  IsItemIndexIconified(&windows, i));
-	redraw = 1;
-      }
+      if (DeskOnly && Desk != DeskNumber)
+	break;
+      if (!GetItemGeometry(&windows, i, &pold_g))
+	break;
+      if (PageOnly && !fvwmrect_do_rectangles_intersect(pold_g, &global_scr_g))
+	break;
+      if (ScreenOnly && !fvwmrect_do_rectangles_intersect(pold_g, &screen_g))
+	break;
+      if (UseSkipList && IsItemIndexSkipWindowList(&windows, i))
+	break;
+      AddButton(&buttons, string, NULL, BUTTON_UP, i,
+		IsItemIndexIconified(&windows, i));
+      redraw = 1;
     }
     else
       redraw = 0;
@@ -741,6 +851,12 @@ void ProcessMessage(unsigned long type,unsigned long *body)
 	  redraw = 1;
 	}
       }
+      else if (strncasecmp(tline, XINERAMA_CONFIG_STRING,
+			   sizeof(XINERAMA_CONFIG_STRING) - 1) == 0)
+      {
+	FScreenConfigureModule(
+	  tline + sizeof(XINERAMA_CONFIG_STRING) - 1);
+      }
     }
     break;
   } /* switch */
@@ -839,6 +955,7 @@ static char *configopts[] =
 {
   "imagepath",
   "colorset",
+  XINERAMA_CONFIG_STRING,
   NULL
 };
 
@@ -871,6 +988,8 @@ static char *moduleopts[] =
   "FocusColorset",
   "3DFvwm",
   "Rows",
+  "PageOnly",
+  "ScreenOnly",
   NULL
 };
 
@@ -902,11 +1021,22 @@ static void ParseConfigLine(char *tline)
     switch(index)
     {
     case 0: /* imagepath */
+      if (ImagePath)
+      {
+	free(ImagePath);
+      }
       CopyString(&ImagePath, rest);
       break;
     case 1: /* colorset */
       LoadColorset(rest);
       break;
+    case 2: /* XINERAMA_CONFIG_STRING */
+      if (XineramaConfig)
+      {
+	free(XineramaConfig);
+      }
+      CopyString(&XineramaConfig, rest);
+      break;
     default:
       /* unknown option */
       break;
@@ -973,6 +1103,8 @@ static void ParseConfigLine(char *tline)
     case 13: /* AutoHide */
       AutoHide=True;
       AutoStick=True;
+      VisiblePixels=atoi(rest);
+      if (VisiblePixels < 1) VisiblePixels = DEFAULT_VISIBLE_PIXELS;
       break;
     case 14: /* UseIconNames */
       UseIconNames=True;
@@ -1022,6 +1154,12 @@ static void ParseConfigLine(char *tline)
 	RowsNumber = 1;
       }
       break;
+    case 27: /* PageOnly */
+      PageOnly=True;
+      break;
+    case 28: /* ScreenOnly */
+      ScreenOnly=True;
+      break;
     default:
       if (!GoodiesParseConfig(tline) &&
 	  !StartButtonParseConfig(tline))
@@ -1208,7 +1346,7 @@ void LoopOnEvents(void)
 	  y = win_y + RowHeight;
 	} else {
 	  /* bar in bottom of the screen */
-	  y = win_y - ScreenHeight;
+	  y = win_y - screen_g.height;
 	}
 	sprintf(tmp,"Popup %s %d %d", StartPopup, x, y);
 	SendFvwmPipe(Fvwm_fd, tmp, 0);
@@ -1228,6 +1366,12 @@ void LoopOnEvents(void)
       break;
 
     case Expose:
+      memcpy(&Event2, &Event, sizeof(Event));
+      /* eat up excess Expose events. */
+      while (XCheckTypedWindowEvent(dpy, win, Expose, &Event2))
+      {
+	memcpy(&Event, &Event2, sizeof(Event));
+      }
       if (Event.xexpose.count == 0)
       {
 	if (Event.xexpose.window == Tip.win)
@@ -1333,22 +1477,6 @@ void LoopOnEvents(void)
       }
       if (Event.xconfigure.height != win_height) {
 	AdjustWindow(Event.xconfigure.width, Event.xconfigure.height);
-	if (AutoHide && !win_is_shaded)
-	{
-	  if (win_y > Midline)
-	    win_y = ScreenHeight - 2 +
-	      win_title_height*win_has_bottom_title;
-	  else
-	    win_y = 2 + win_title_height*win_has_bottom_title
-	      - win_height;
-	  XSync(dpy,0);
-	  XMoveWindow(dpy, win, win_x, win_y);
-	  XSync(dpy,0);
-	  hide_taskbar_alarm = False;
-	  WindowState = -1;
-	}
-	else if (AutoStick)
-	  WarpTaskBar(win_y, 1);
 	redraw = 1;
       }
       /* useful because of dynamic style change */
@@ -1357,8 +1485,6 @@ void LoopOnEvents(void)
 	PurgeConfigEvents();
 	break;
       }
-      else if (AutoHide)
-	break;
       else if (Event.xconfigure.x != win_x || Event.xconfigure.y != win_y)
       {
 	if (AutoStick)
@@ -1645,7 +1771,9 @@ static Bool change_colorset(int cset, Bool force)
 void StartMeUp(void)
 {
    XSizeHints hints;
+   XClassHint classhints;
    int ret;
+   int i;
    XSetWindowAttributes attr;
 #ifdef I18N_MB
   char **ml;
@@ -1655,29 +1783,47 @@ void StartMeUp(void)
 #endif
 
    if (!(dpy = XOpenDisplay(""))) {
-      fprintf(stderr,"%s: can't open display %s", Module,
-	      XDisplayName(""));
-      exit (1);
+     fprintf(stderr,"%s: can't open display %s", Module,
+	     XDisplayName(""));
+     exit (1);
    }
    InitPictureCMap(dpy);
+   FScreenInit(dpy);
+   if (XineramaConfig)
+   {
+     FScreenConfigureModule(XineramaConfig);
+     free(XineramaConfig);
+   }
    AllocColorset(0);
+   FShapeInit(dpy);
    x_fd = XConnectionNumber(dpy);
    screen= DefaultScreen(dpy);
    Root = RootWindow(dpy, screen);
 
-   ScreenWidth  = XDisplayWidth(dpy, screen);
-   ScreenHeight = XDisplayHeight(dpy, screen);
+   if (geometry == NULL)
+     UpdateString(&geometry, "+0-0");
+   /* evaluate further down */
+   ret = FScreenParseGeometryWithScreen(
+     geometry, &hints.x, &hints.y, (unsigned int *)&hints.width,
+     (unsigned int *)&hints.height, &fscreen);
+   FScreenGetScrRect(
+     NULL, fscreen,
+     &screen_g.x, &screen_g.y, &screen_g.width, &screen_g.height);
+   FScreenGetScrRect(
+     NULL, FSCREEN_GLOBAL,
+     &global_scr_g.x, &global_scr_g.y, &global_scr_g.width,
+     &global_scr_g.height);
+   Midline = (screen_g.height >> 1) + screen_g.y;
 
-   Midline = (int) (ScreenHeight >> 1);
-
-   if (selfont_string == NULL) selfont_string = font_string;
+   if (selfont_string == NULL)
+     selfont_string = font_string;
 
 #ifdef I18N_MB
    if ((ButtonFontset=XCreateFontSet(dpy,font_string,&ml,&mc,&ds)) == NULL) {
 #ifdef STRICTLY_FIXED
      if ((ButtonFontset=XCreateFontSet(dpy,"fixed",&ml,&mc,&ds)) == NULL) {
 #else
-     if ((ButtonFontset=XCreateFontSet(dpy,"-*-fixed-medium-r-normal-*-14-*-*-*-*-*-*-*",&ml,&mc,&ds)) == NULL) { 
+     if ((ButtonFontset=XCreateFontSet(dpy,"-*-fixed-medium-r-normal-*-14-*-*-*-*-*-*-*",&ml,&mc,&ds)) == NULL) {
 #endif
        fprintf(stderr, "%s: Couldn't load fixed font. Exiting!\n",Module);
        exit(1);
@@ -1685,7 +1831,8 @@ void StartMeUp(void)
    }
    XFontsOfFontSet(ButtonFontset,&fs_list,&ml);
    ButtonFont = fs_list[0];
-   if ((SelButtonFontset=XCreateFontSet(dpy,selfont_string,&ml,&mc,&ds)) == NULL) {
+   if ((SelButtonFontset = XCreateFontSet(dpy,selfont_string,&ml,&mc,&ds))
+       == NULL) {
 #ifdef STRICTLY_FIXED
      if ((SelButtonFontset=XCreateFontSet(dpy,"fixed",&ml,&mc,&ds)) == NULL)
 #else
@@ -1718,48 +1865,51 @@ void StartMeUp(void)
 
    win_border = 4; /* default border width */
    win_height = RowHeight+(RowsNumber-1)*(RowHeight+2);
-   win_width = ScreenWidth - (win_border * 2);
-
-   if (geometry == NULL)
-     UpdateString(&geometry, "+0-0");
-   ret = XParseGeometry(geometry, &hints.x, &hints.y,
-	 	        (unsigned int *)&hints.width,
-		        (unsigned int *)&hints.height);
+   win_width = screen_g.width - (win_border * 2);
 
+   /* now evaluate the geometry parsed above */
    if (ret & YNegative)
    {
      if (AutoStick)
      {
        if (-hints.y < Midline)
-	 hints.y = ScreenHeight - (AutoHide ? 2 : win_height);
+	 hints.y = screen_g.height - (AutoHide ? VISIBLE_PIXELS() : win_height);
        else
-	 hints.y = AutoHide ? 2 - win_height : 0;
+	 hints.y = AutoHide ? VISIBLE_PIXELS() - win_height : 0;
+       hints.y += screen_g.y;
      }
      else
-       hints.y += ScreenHeight - win_height;
+     {
+       hints.y += screen_g.height - win_height;
+     }
    }
    else if (AutoStick)
    {
      if (hints.y < Midline)
-       hints.y = AutoHide ? 2 - win_height : 0;
+       hints.y = AutoHide ? VISIBLE_PIXELS() - win_height : 0;
      else
-       hints.y = ScreenHeight - (AutoHide ? 2 : win_height);
+       hints.y = screen_g.height - (AutoHide ? VISIBLE_PIXELS() : win_height);
+     hints.y += screen_g.y;
    }
 
-    if (ret & XNegative)
-    {
-      if (ret & YNegative) hints.win_gravity=SouthEastGravity;
-      else hints.win_gravity=NorthEastGravity;
-    }
-    else
-    {
-      if (ret & YNegative) hints.win_gravity=SouthWestGravity;
-      else hints.win_gravity=NorthWestGravity;
-    }
+   if (ret & XNegative)
+   {
+     if (ret & YNegative)
+       hints.win_gravity=SouthEastGravity;
+     else
+       hints.win_gravity=NorthEastGravity;
+   }
+   else
+   {
+     if (ret & YNegative)
+       hints.win_gravity=SouthWestGravity;
+     else
+       hints.win_gravity=NorthWestGravity;
+   }
 
    hints.flags=USPosition|PPosition|USSize|PSize|PResizeInc|
      PWinGravity|PMinSize|PMaxSize|PBaseSize;
-   hints.x           = 0;
+   hints.x           = screen_g.x;
    hints.width       = win_width;
    hints.height      = win_height;
    hints.width_inc   = 1;
@@ -1807,14 +1957,29 @@ void StartMeUp(void)
    wm_del_win=XInternAtom(dpy,"WM_DELETE_WINDOW",False);
    XSetWMProtocols(dpy,win,&wm_del_win,1);
 
-   XSetWMNormalHints(dpy,win,&hints);
+  {
+    XTextProperty nametext;
+    char *list[]={NULL,NULL};
+    list[0] = Module+1;
 
-   XGrabButton(dpy,1,AnyModifier,win,True,GRAB_EVENTS,GrabModeAsync,
-	       GrabModeAsync,None,None);
-   XGrabButton(dpy,2,AnyModifier,win,True,GRAB_EVENTS,GrabModeAsync,
-	       GrabModeAsync,None,None);
-   XGrabButton(dpy,3,AnyModifier,win,True,GRAB_EVENTS,GrabModeAsync,
-	       GrabModeAsync,None,None);
+    classhints.res_name=safestrdup(Module+1);
+    classhints.res_class=safestrdup("FvwmTaskBar");
+
+    if(!XStringListToTextProperty(list,1,&nametext))
+    {
+      fprintf(stderr,"%s: Failed to convert name to XText\n",Module);
+      exit(1);
+    }
+    XSetWMProperties(dpy,win,&nametext,&nametext,
+		     NULL,0,&hints,NULL,&classhints);
+    XFree(nametext.value);
+  }
+
+   for (i = 1; i <= NUMBER_OF_MOUSE_BUTTONS; i++)
+   {
+     XGrabButton(dpy,i,AnyModifier,win,True,GRAB_EVENTS,GrabModeAsync,
+		 GrabModeAsync,None,None);
+   }
 
    /*   SetMwmHints(MWM_DECOR_ALL|MWM_DECOR_MAXIMIZE|MWM_DECOR_MINIMIZE,
 	       MWM_FUNC_ALL|MWM_FUNC_MAXIMIZE|MWM_FUNC_MINIMIZE,
@@ -1833,8 +1998,6 @@ void StartMeUp(void)
                          EnterWindowMask | LeaveWindowMask |
                          StructureNotifyMask));
 
-   ChangeWindowName(&Module[1]);
-
    InitGoodies();
    atexit(ShutMeDown);
 }
@@ -1918,12 +2081,18 @@ PropMwmHints prop;
 void WarpTaskBar(int y, Bool force)
 {
   /* The tests on y are really useful ! */
-  if (!AutoHide && ((y != (int)ScreenHeight - win_height - win_border &&
-      y !=  win_border) || force)) {
+  if (!AutoHide &&
+      ((y != screen_g.y + screen_g.height - win_height - win_border &&
+	y !=  screen_g.y + win_border) || force)) {
     if (y > Midline)
-      win_y = (int)ScreenHeight - win_height - win_border;
+    {
+      win_y = screen_g.height - win_height - win_border;
+    }
     else
+    {
       win_y = win_border + win_title_height;
+    }
+    win_y += screen_g.y;
     XSync(dpy, 0);
     XMoveWindow(dpy, win, win_x, win_y);
     XSync(dpy, 0);
@@ -1939,6 +2108,23 @@ void WarpTaskBar(int y, Bool force)
   PurgeConfigEvents();
 }
 
+/***********************************************************************
+ SleepALittle -- do a small delay in a fairly portable way
+ Replace "#if 1" with "#if 0" to turn delay off
+ ***********************************************************************/
+static void SleepALittle(void)
+{
+  struct timeval tv;
+
+    tv.tv_sec  = 0;
+    tv.tv_usec = 10000;
+#if 1
+    XSync(dpy, 0);
+    select(0, NULL, NULL, NULL, &tv);
+    XSync(dpy, 0);
+#endif
+}
+
 /***********************************************************************
  RevealTaskBar -- Make taskbar fully visible
  ***********************************************************************/
@@ -1955,14 +2141,23 @@ void RevealTaskBar()
   /* Go faster with the number of rows */
   inc_y += (NRows >= 3 ? 3 : 0) + (NRows >= 5 ? 3 : 0);
 
-  if (win_y < Midline) {
-    new_win_y = win_border + win_title_height;
+  if (win_y < Midline)
+  {
+    new_win_y = screen_g.y + win_border + win_title_height;
     for (; win_y<=new_win_y; win_y +=inc_y)
+    {
       XMoveWindow(dpy, win, win_x, win_y);
-  } else {
-    new_win_y = (int)ScreenHeight - win_height - win_border;
+      SleepALittle();
+    }
+  }
+  else
+  {
+    new_win_y = screen_g.y + screen_g.height - win_height - win_border;
     for (; win_y>=new_win_y; win_y -=inc_y)
+    {
       XMoveWindow(dpy, win, win_x, win_y);
+      SleepALittle();
+    }
   }
 
   win_y = new_win_y;
@@ -2003,16 +2198,25 @@ void HideTaskBar()
   /* Go faster with the number of rows */
   inc_y += (NRows >= 3 ? 2 : 0) + (NRows >= 5 ? 2 : 0);
 
-  if (win_y < Midline) {
-    new_win_y = 2 + win_title_height*win_has_bottom_title -
-      win_height;
+  if (win_y < Midline)
+  {
+    new_win_y = screen_g.y + VISIBLE_PIXELS() +
+      win_title_height*win_has_bottom_title - win_height;
     for (; win_y>=new_win_y; win_y -=inc_y)
+    {
       XMoveWindow(dpy, win, win_x, win_y);
-  } else {
-    new_win_y = (int)ScreenHeight - 2 +
+      SleepALittle();
+    }
+  }
+  else
+  {
+    new_win_y = screen_g.y + screen_g.height - VISIBLE_PIXELS() +
       win_title_height*win_has_bottom_title;
     for (; win_y<=new_win_y; win_y +=inc_y)
+    {
       XMoveWindow(dpy, win, win_x, win_y);
+      SleepALittle();
+    }
   }
   win_y = new_win_y;
   XMoveWindow(dpy, win, win_x, win_y);
